package collection

import (
	"fmt"
	"strings"
	"sync"

	"github.com/anyproto/any-sync/app"
	"github.com/gogo/protobuf/types"
	"github.com/samber/lo"

	"github.com/anyproto/anytype-heart/core/block"
	"github.com/anyproto/anytype-heart/core/block/editor/basic"
	"github.com/anyproto/anytype-heart/core/block/editor/smartblock"
	"github.com/anyproto/anytype-heart/core/block/editor/state"
	"github.com/anyproto/anytype-heart/core/block/editor/template"
	"github.com/anyproto/anytype-heart/core/domain"
	"github.com/anyproto/anytype-heart/core/session"
	"github.com/anyproto/anytype-heart/pb"
	"github.com/anyproto/anytype-heart/pkg/lib/bundle"
	coresb "github.com/anyproto/anytype-heart/pkg/lib/core/smartblock"
	"github.com/anyproto/anytype-heart/pkg/lib/localstore/addr"
	"github.com/anyproto/anytype-heart/pkg/lib/localstore/objectstore"
	"github.com/anyproto/anytype-heart/pkg/lib/logging"
	"github.com/anyproto/anytype-heart/pkg/lib/pb/model"
	"github.com/anyproto/anytype-heart/util/internalflag"
	"github.com/anyproto/anytype-heart/util/pbtypes"
	"github.com/anyproto/anytype-heart/util/slice"
)

var log = logging.Logger("collection-service")

type Service struct {
	lock        *sync.RWMutex
	collections map[string]map[string]chan []string
	picker      block.ObjectGetter
	objectStore objectstore.ObjectStore
}

func New() *Service {
	return &Service{
		lock:        &sync.RWMutex{},
		collections: map[string]map[string]chan []string{},
	}
}

func (s *Service) Init(a *app.App) (err error) {
	s.picker = app.MustComponent[block.ObjectGetter](a)
	s.objectStore = app.MustComponent[objectstore.ObjectStore](a)
	return nil
}

func (s *Service) Name() string {
	return "collection"
}

func (s *Service) CollectionType() string {
	return "collection"
}

func (s *Service) Add(ctx session.Context, req *pb.RpcObjectCollectionAddRequest) error {
	return s.updateCollection(ctx, req.ContextId, func(col []string) []string {
		toAdd := slice.Difference(req.ObjectIds, col)
		pos := slice.FindPos(col, req.AfterId)
		if pos >= 0 {
			col = slice.Insert(col, pos+1, toAdd...)
		} else {
			col = append(toAdd, col...)
		}
		return col
	})
}

func (s *Service) Remove(ctx session.Context, req *pb.RpcObjectCollectionRemoveRequest) error {
	return s.updateCollection(ctx, req.ContextId, func(col []string) []string {
		col = slice.Filter(col, func(id string) bool {
			return slice.FindPos(req.ObjectIds, id) == -1
		})
		return col
	})
}

func (s *Service) Sort(ctx session.Context, req *pb.RpcObjectCollectionSortRequest) error {
	return s.updateCollection(ctx, req.ContextId, func(col []string) []string {
		exist := map[string]struct{}{}
		for _, id := range col {
			exist[id] = struct{}{}
		}
		col = col[:0]
		for _, id := range req.ObjectIds {
			// Reorder only existing objects
			if _, ok := exist[id]; ok {
				col = append(col, id)
			}
		}
		return col
	})
}

func (s *Service) updateCollection(ctx session.Context, contextID string, modifier func(src []string) []string) error {
	return block.DoStateCtx(s.picker, ctx, contextID, func(s *state.State, sb smartblock.SmartBlock) error {
		lst := s.GetStoreSlice(template.CollectionStoreKey)
		lst = modifier(lst)
		s.UpdateStoreSlice(template.CollectionStoreKey, lst)
		return nil
	})
}

func (s *Service) collectionAddHookOnce(sb smartblock.SmartBlock) {
	sb.AddHookOnce("collection", func(info smartblock.ApplyInfo) (err error) {
		for _, ch := range info.Changes {
			if upd := ch.GetStoreSliceUpdate(); upd != nil && upd.Key == template.CollectionStoreKey {
				s.broadcast(sb.Id(), info.State.GetStoreSlice(template.CollectionStoreKey))
				return nil
			}
		}
		return nil
	}, smartblock.HookAfterApply)
}

func (s *Service) broadcast(collectionID string, objectIDs []string) {
	s.lock.RLock()
	defer s.lock.RUnlock()

	for _, ch := range s.collections[collectionID] {
		ch <- objectIDs
	}
}

type Subscription struct {
	objectsCh chan []string
	closeCh   chan struct{}
}

func (s *Subscription) Chan() <-chan []string {
	return s.objectsCh
}

func (s *Subscription) Close() {
	close(s.closeCh)
}

func (s *Service) SubscribeForCollection(collectionID string, subscriptionID string) ([]string, <-chan []string, error) {
	var initialObjectIDs []string

	s.lock.Lock()
	defer s.lock.Unlock()

	col, ok := s.collections[collectionID]
	if !ok {
		col = map[string]chan []string{}
		s.collections[collectionID] = col
	}
	err := block.DoStateAsync(s.picker, collectionID, func(st *state.State, sb smartblock.SmartBlock) error {
		s.collectionAddHookOnce(sb)

		initialObjectIDs = st.GetStoreSlice(template.CollectionStoreKey)
		return nil
	})
	if err != nil {
		return nil, nil, err
	}

	ch, ok := col[subscriptionID]
	if !ok {
		ch = make(chan []string)
		col[subscriptionID] = ch
	}

	return initialObjectIDs, ch, err
}

func (s *Service) UnsubscribeFromCollection(collectionID string, subscriptionID string) {
	s.lock.Lock()
	defer s.lock.Unlock()

	col, ok := s.collections[collectionID]
	if !ok {
		return
	}

	ch, ok := col[subscriptionID]
	if ok {
		close(ch)
		delete(col, subscriptionID)
	}
}

func (s *Service) CreateCollection(details *types.Struct, flags []*model.InternalFlag) (coresb.SmartBlockType, *types.Struct, *state.State, error) {
	details = internalflag.PutToDetails(details, flags)

	newState := state.NewDoc("", nil).NewState().SetDetails(details)

	tmpls := []template.StateTransformer{
		template.WithRequiredRelations(),
	}

	blockContent := template.MakeCollectionDataviewContent()
	tmpls = append(tmpls,
		template.WithDataview(blockContent, false),
	)
	template.InitTemplate(newState, tmpls...)

	return coresb.SmartBlockTypePage, newState.CombinedDetails(), newState, nil
}

func (s *Service) ObjectToCollection(id string) error {
	if err := block.Do(s.picker, id, func(b smartblock.SmartBlock) error {
		commonOperations, ok := b.(basic.CommonOperations)
		if !ok {
			return fmt.Errorf("invalid smartblock impmlementation: %T", b)
		}
		st := b.NewState()
		err := commonOperations.SetLayoutInStateAndIgnoreRestriction(st, model.ObjectType_collection)
		if err != nil {
			return fmt.Errorf("set layout: %w", err)
		}
		st.SetObjectTypeKey(bundle.TypeKeyCollection)
		setDefaultObjectTypeToViews(st)
		flags := internalflag.NewFromState(st)
		flags.Remove(model.InternalFlag_editorSelectType)
		flags.Remove(model.InternalFlag_editorDeleteEmpty)
		flags.AddToState(st)
		return b.Apply(st)
	}); err != nil {
		return err
	}

	return nil
}

func setDefaultObjectTypeToViews(st *state.State) {
	if !lo.Contains(st.ObjectTypeKeys(), bundle.TypeKeySet) {
		return
	}

	setOfValue := pbtypes.GetStringList(st.ParentState().Details(), bundle.RelationKeySetOf.String())
	if len(setOfValue) == 0 || !strings.HasPrefix(setOfValue[0], addr.ObjectTypeKeyToIdPrefix) {
		return
	}

	if isNotCreatableType(domain.TypeKey(strings.TrimPrefix(setOfValue[0], addr.ObjectTypeKeyToIdPrefix))) {
		return
	}

	dataviewBlock := st.Get(state.DataviewBlockID)
	if dataviewBlock == nil {
		return
	}
	content, ok := dataviewBlock.Model().Content.(*model.BlockContentOfDataview)
	if !ok {
		return
	}

	for _, view := range content.Dataview.Views {
		view.DefaultObjectTypeId = setOfValue[0]
	}
}

func isNotCreatableType(key domain.TypeKey) bool {
	return lo.Contains(append(bundle.InternalTypes, bundle.TypeKeyObjectType), key)
}
